@@ -51,7 +51,7 @@ static void checkfutureword(JF, js_Ast *exp)
 	}
 }
 
-static js_Function *newfun(js_State *J, js_Ast *name, js_Ast *params, js_Ast *body, int script, int default_strict)
+static js_Function *newfun(js_State *J, int line, js_Ast *name, js_Ast *params, js_Ast *body, int script, int default_strict)
 {
 	js_Function *F = js_malloc(J, sizeof *F);
 	memset(F, 0, sizeof *F);
@@ -61,7 +61,7 @@ static js_Function *newfun(js_State *J, js_Ast *name, js_Ast *params, js_Ast *bo
 	++J->gccounter;
 
 	F->filename = js_intern(J, J->filename);
-	F->line = name ? name->line : params ? params->line : body ? body->line : 1;
+	F->line = line;
 	F->script = script;
 	F->strict = default_strict;
 	F->name = name ? name->string : "";
@@ -85,17 +85,19 @@ static void emitraw(JF, int value)
 }
 
 static void emit(JF, int value)
+{
+	emitraw(J, F, F->lastline);
+	emitraw(J, F, value);
+}
+
+static void emitarg(JF, int value)
 {
 	emitraw(J, F, value);
 }
 
 static void emitline(JF, js_Ast *node)
 {
-	if (F->lastline != node->line) {
-		F->lastline = node->line;
-		emit(J, F, OP_LINE);
-		emitraw(J, F, node->line);
-	}
+	F->lastline = node->line;
 }
 
 static int addfunction(JF, js_Function *value)
@@ -136,7 +138,7 @@ static int addstring(JF, const char *value)
 	return F->strlen++;
 }
 
-static void addlocal(JF, js_Ast *ident, int reuse)
+static int addlocal(JF, js_Ast *ident, int reuse)
 {
 	const char *name = ident->string;
 	if (F->strict) {
@@ -144,13 +146,16 @@ static void addlocal(JF, js_Ast *ident, int reuse)
 			jsC_error(J, ident, "redefining 'arguments' is not allowed in strict mode");
 		if (!strcmp(name, "eval"))
 			jsC_error(J, ident, "redefining 'eval' is not allowed in strict mode");
+	} else {
+		if (!strcmp(name, "eval"))
+			js_evalerror(J, "%s:%d: invalid use of 'eval'", J->filename, ident->line);
 	}
 	if (reuse || F->strict) {
 		int i;
 		for (i = 0; i < F->varlen; ++i) {
 			if (!strcmp(F->vartab[i], name)) {
 				if (reuse)
-					return;
+					return i+1;
 				if (F->strict)
 					jsC_error(J, ident, "duplicate formal parameter '%s'", name);
 			}
@@ -160,7 +165,8 @@ static void addlocal(JF, js_Ast *ident, int reuse)
 		F->varcap = F->varcap ? F->varcap * 2 : 16;
 		F->vartab = js_realloc(J, F->vartab, F->varcap * sizeof *F->vartab);
 	}
-	F->vartab[F->varlen++] = name;
+	F->vartab[F->varlen] = name;
+	return ++F->varlen;
 }
 
 static int findlocal(JF, const char *name)
@@ -174,55 +180,61 @@ static int findlocal(JF, const char *name)
 
 static void emitfunction(JF, js_Function *fun)
 {
+	F->lightweight = 0;
 	emit(J, F, OP_CLOSURE);
-	emitraw(J, F, addfunction(J, F, fun));
+	emitarg(J, F, addfunction(J, F, fun));
 }
 
 static void emitnumber(JF, double num)
 {
 	if (num == 0) {
-		emit(J, F, OP_NUMBER_0);
+		emit(J, F, OP_INTEGER);
+		emitarg(J, F, 32768);
 		if (signbit(num))
 			emit(J, F, OP_NEG);
-	} else if (num == 1) {
-		emit(J, F, OP_NUMBER_1);
-	} else if (num == (js_Instruction)num) {
-		emit(J, F, OP_NUMBER_POS);
-		emitraw(J, F, (js_Instruction)num);
-	} else if (num < 0 && -num == (js_Instruction)(-num)) {
-		emit(J, F, OP_NUMBER_NEG);
-		emitraw(J, F, (js_Instruction)(-num));
+	} else if (num >= SHRT_MIN && num <= SHRT_MAX && num == (int)num) {
+		emit(J, F, OP_INTEGER);
+		emitarg(J, F, num + 32768);
 	} else {
 		emit(J, F, OP_NUMBER);
-		emitraw(J, F, addnumber(J, F, num));
+		emitarg(J, F, addnumber(J, F, num));
 	}
 }
 
 static void emitstring(JF, int opcode, const char *str)
 {
 	emit(J, F, opcode);
-	emitraw(J, F, addstring(J, F, str));
+	emitarg(J, F, addstring(J, F, str));
 }
 
 static void emitlocal(JF, int oploc, int opvar, js_Ast *ident)
 {
+	int is_arguments = !strcmp(ident->string, "arguments");
+	int is_eval = !strcmp(ident->string, "eval");
 	int i;
+
+	if (is_arguments) {
+		F->lightweight = 0;
+		F->arguments = 1;
+	}
+
 	checkfutureword(J, F, ident);
 	if (F->strict && oploc == OP_SETLOCAL) {
-		if (!strcmp(ident->string, "arguments"))
+		if (is_arguments)
 			jsC_error(J, ident, "'arguments' is read-only in strict mode");
-		if (!strcmp(ident->string, "eval"))
+		if (is_eval)
 			jsC_error(J, ident, "'eval' is read-only in strict mode");
 	}
-	if (F->lightweight) {
-		i = findlocal(J, F, ident->string);
-		if (i >= 0) {
-			emit(J, F, oploc);
-			emitraw(J, F, i);
-			return;
-		}
+	if (is_eval)
+		js_evalerror(J, "%s:%d: invalid use of 'eval'", J->filename, ident->line);
+
+	i = findlocal(J, F, ident->string);
+	if (i < 0) {
+		emitstring(J, F, opvar, ident->string);
+	} else {
+		emit(J, F, oploc);
+		emitarg(J, F, i);
 	}
-	emitstring(J, F, opvar, ident->string);
 }
 
 static int here(JF)
@@ -232,9 +244,10 @@ static int here(JF)
 
 static int emitjump(JF, int opcode)
 {
-	int inst = F->codelen + 1;
+	int inst;
 	emit(J, F, opcode);
-	emitraw(J, F, 0);
+	inst = F->codelen;
+	emitarg(J, F, 0);
 	return inst;
 }
 
@@ -243,7 +256,7 @@ static void emitjumpto(JF, int opcode, int dest)
 	emit(J, F, opcode);
 	if (dest != (js_Instruction)dest)
 		js_syntaxerror(J, "jump address integer overflow");
-	emitraw(J, F, dest);
+	emitarg(J, F, dest);
 }
 
 static void labelto(JF, int inst, int addr)
@@ -262,16 +275,20 @@ static void label(JF, int inst)
 
 static void ctypeof(JF, js_Ast *exp)
 {
-	if (exp->type == EXP_IDENTIFIER)
-		emitlocal(J, F, OP_GETLOCAL, OP_HASVAR, exp);
-	else
-		cexp(J, F, exp);
+	if (exp->a->type == EXP_IDENTIFIER) {
+		emitline(J, F, exp->a);
+		emitlocal(J, F, OP_GETLOCAL, OP_HASVAR, exp->a);
+	} else {
+		cexp(J, F, exp->a);
+	}
+	emitline(J, F, exp);
 	emit(J, F, OP_TYPEOF);
 }
 
 static void cunary(JF, js_Ast *exp, int opcode)
 {
 	cexp(J, F, exp->a);
+	emitline(J, F, exp);
 	emit(J, F, opcode);
 }
 
@@ -279,6 +296,7 @@ static void cbinary(JF, js_Ast *exp, int opcode)
 {
 	cexp(J, F, exp->a);
 	cexp(J, F, exp->b);
+	emitline(J, F, exp);
 	emit(J, F, opcode);
 }
 
@@ -287,9 +305,10 @@ static void carray(JF, js_Ast *list)
 	int i = 0;
 	while (list) {
 		if (list->a->type != EXP_UNDEF) {
-			emitnumber(J, F, i++);
 			emitline(J, F, list->a);
+			emitnumber(J, F, i++);
 			cexp(J, F, list->a);
+			emitline(J, F, list->a);
 			emit(J, F, OP_INITPROP);
 		} else {
 			++i;
@@ -330,12 +349,15 @@ static void cobject(JF, js_Ast *list)
 		js_Ast *kv = list->a;
 		js_Ast *prop = kv->a;
 
-		if (prop->type == AST_IDENTIFIER || prop->type == EXP_STRING)
+		if (prop->type == AST_IDENTIFIER || prop->type == EXP_STRING) {
+			emitline(J, F, prop);
 			emitstring(J, F, OP_STRING, prop->string);
-		else if (prop->type == EXP_NUMBER)
+		} else if (prop->type == EXP_NUMBER) {
+			emitline(J, F, prop);
 			emitnumber(J, F, prop->number);
-		else
+		} else {
 			jsC_error(J, prop, "invalid property name in object initializer");
+		}
 
 		if (F->strict)
 			checkdup(J, F, head, kv);
@@ -343,16 +365,18 @@ static void cobject(JF, js_Ast *list)
 		switch (kv->type) {
 		default: /* impossible */ break;
 		case EXP_PROP_VAL:
-			emitline(J, F, kv->b);
 			cexp(J, F, kv->b);
+			emitline(J, F, kv);
 			emit(J, F, OP_INITPROP);
 			break;
 		case EXP_PROP_GET:
-			emitfunction(J, F, newfun(J, NULL, NULL, kv->c, 0, F->strict));
+			emitfunction(J, F, newfun(J, prop->line, NULL, NULL, kv->c, 0, F->strict));
+			emitline(J, F, kv);
 			emit(J, F, OP_INITGETTER);
 			break;
 		case EXP_PROP_SET:
-			emitfunction(J, F, newfun(J, NULL, kv->b, kv->c, 0, F->strict));
+			emitfunction(J, F, newfun(J, prop->line, NULL, kv->b, kv->c, 0, F->strict));
+			emitline(J, F, kv);
 			emit(J, F, OP_INITSETTER);
 			break;
 		}
@@ -379,17 +403,20 @@ static void cassign(JF, js_Ast *exp)
 	switch (lhs->type) {
 	case EXP_IDENTIFIER:
 		cexp(J, F, rhs);
+		emitline(J, F, exp);
 		emitlocal(J, F, OP_SETLOCAL, OP_SETVAR, lhs);
 		break;
 	case EXP_INDEX:
 		cexp(J, F, lhs->a);
 		cexp(J, F, lhs->b);
 		cexp(J, F, rhs);
+		emitline(J, F, exp);
 		emit(J, F, OP_SETPROP);
 		break;
 	case EXP_MEMBER:
 		cexp(J, F, lhs->a);
 		cexp(J, F, rhs);
+		emitline(J, F, exp);
 		emitstring(J, F, OP_SETPROP_S, lhs->b->string);
 		break;
 	default:
@@ -404,6 +431,7 @@ static void cassignforin(JF, js_Ast *stm)
 	if (stm->type == STM_FOR_IN_VAR) {
 		if (lhs->b)
 			jsC_error(J, lhs->b, "more than one loop variable in for-in statement");
+		emitline(J, F, lhs->a);
 		emitlocal(J, F, OP_SETLOCAL, OP_SETVAR, lhs->a->a); /* list(var-init(ident)) */
 		emit(J, F, OP_POP);
 		return;
@@ -411,18 +439,21 @@ static void cassignforin(JF, js_Ast *stm)
 
 	switch (lhs->type) {
 	case EXP_IDENTIFIER:
+		emitline(J, F, lhs);
 		emitlocal(J, F, OP_SETLOCAL, OP_SETVAR, lhs);
 		emit(J, F, OP_POP);
 		break;
 	case EXP_INDEX:
 		cexp(J, F, lhs->a);
 		cexp(J, F, lhs->b);
+		emitline(J, F, lhs);
 		emit(J, F, OP_ROT3);
 		emit(J, F, OP_SETPROP);
 		emit(J, F, OP_POP);
 		break;
 	case EXP_MEMBER:
 		cexp(J, F, lhs->a);
+		emitline(J, F, lhs);
 		emit(J, F, OP_ROT2);
 		emitstring(J, F, OP_SETPROP_S, lhs->b->string);
 		emit(J, F, OP_POP);
@@ -436,16 +467,19 @@ static void cassignop1(JF, js_Ast *lhs)
 {
 	switch (lhs->type) {
 	case EXP_IDENTIFIER:
+		emitline(J, F, lhs);
 		emitlocal(J, F, OP_GETLOCAL, OP_GETVAR, lhs);
 		break;
 	case EXP_INDEX:
 		cexp(J, F, lhs->a);
 		cexp(J, F, lhs->b);
+		emitline(J, F, lhs);
 		emit(J, F, OP_DUP2);
 		emit(J, F, OP_GETPROP);
 		break;
 	case EXP_MEMBER:
 		cexp(J, F, lhs->a);
+		emitline(J, F, lhs);
 		emit(J, F, OP_DUP);
 		emitstring(J, F, OP_GETPROP_S, lhs->b->string);
 		break;
@@ -458,14 +492,17 @@ static void cassignop2(JF, js_Ast *lhs, int postfix)
 {
 	switch (lhs->type) {
 	case EXP_IDENTIFIER:
+		emitline(J, F, lhs);
 		if (postfix) emit(J, F, OP_ROT2);
 		emitlocal(J, F, OP_SETLOCAL, OP_SETVAR, lhs);
 		break;
 	case EXP_INDEX:
+		emitline(J, F, lhs);
 		if (postfix) emit(J, F, OP_ROT4);
 		emit(J, F, OP_SETPROP);
 		break;
 	case EXP_MEMBER:
+		emitline(J, F, lhs);
 		if (postfix) emit(J, F, OP_ROT3);
 		emitstring(J, F, OP_SETPROP_S, lhs->b->string);
 		break;
@@ -480,26 +517,31 @@ static void cassignop(JF, js_Ast *exp, int opcode)
 	js_Ast *rhs = exp->b;
 	cassignop1(J, F, lhs);
 	cexp(J, F, rhs);
+	emitline(J, F, exp);
 	emit(J, F, opcode);
 	cassignop2(J, F, lhs, 0);
 }
 
 static void cdelete(JF, js_Ast *exp)
 {
-	switch (exp->type) {
+	js_Ast *arg = exp->a;
+	switch (arg->type) {
 	case EXP_IDENTIFIER:
 		if (F->strict)
 			jsC_error(J, exp, "delete on an unqualified name is not allowed in strict mode");
-		emitlocal(J, F, OP_DELLOCAL, OP_DELVAR, exp);
+		emitline(J, F, exp);
+		emitlocal(J, F, OP_DELLOCAL, OP_DELVAR, arg);
 		break;
 	case EXP_INDEX:
-		cexp(J, F, exp->a);
-		cexp(J, F, exp->b);
+		cexp(J, F, arg->a);
+		cexp(J, F, arg->b);
+		emitline(J, F, exp);
 		emit(J, F, OP_DELPROP);
 		break;
 	case EXP_MEMBER:
-		cexp(J, F, exp->a);
-		emitstring(J, F, OP_DELPROP_S, exp->b->string);
+		cexp(J, F, arg->a);
+		emitline(J, F, exp);
+		emitstring(J, F, OP_DELPROP_S, arg->b->string);
 		break;
 	default:
 		jsC_error(J, exp, "invalid l-value in delete expression");
@@ -509,6 +551,8 @@ static void cdelete(JF, js_Ast *exp)
 static void ceval(JF, js_Ast *fun, js_Ast *args)
 {
 	int n = cargs(J, F, args);
+	F->lightweight = 0;
+	F->arguments = 1;
 	if (n == 0)
 		emit(J, F, OP_UNDEF);
 	else while (n-- > 1)
@@ -538,7 +582,7 @@ static void ccall(JF, js_Ast *fun, js_Ast *args)
 			ceval(J, F, fun, args);
 			return;
 		}
-		/* fall through */
+		/* fallthrough */
 	default:
 		cexp(J, F, fun);
 		emit(J, F, OP_UNDEF);
@@ -546,7 +590,7 @@ static void ccall(JF, js_Ast *fun, js_Ast *args)
 	}
 	n = cargs(J, F, args);
 	emit(J, F, OP_CALL);
-	emitraw(J, F, n);
+	emitarg(J, F, n);
 }
 
 static void cexp(JF, js_Ast *exp)
@@ -555,46 +599,74 @@ static void cexp(JF, js_Ast *exp)
 	int n;
 
 	switch (exp->type) {
-	case EXP_STRING: emitstring(J, F, OP_STRING, exp->string); break;
-	case EXP_NUMBER: emitnumber(J, F, exp->number); break;
-	case EXP_UNDEF: emit(J, F, OP_UNDEF); break;
-	case EXP_NULL: emit(J, F, OP_NULL); break;
-	case EXP_TRUE: emit(J, F, OP_TRUE); break;
-	case EXP_FALSE: emit(J, F, OP_FALSE); break;
-	case EXP_THIS: emit(J, F, OP_THIS); break;
+	case EXP_STRING:
+		emitline(J, F, exp);
+		emitstring(J, F, OP_STRING, exp->string);
+		break;
+	case EXP_NUMBER:
+		emitline(J, F, exp);
+		emitnumber(J, F, exp->number);
+		break;
+	case EXP_UNDEF:
+		emitline(J, F, exp);
+		emit(J, F, OP_UNDEF);
+		break;
+	case EXP_NULL:
+		emitline(J, F, exp);
+		emit(J, F, OP_NULL);
+		break;
+	case EXP_TRUE:
+		emitline(J, F, exp);
+		emit(J, F, OP_TRUE);
+		break;
+	case EXP_FALSE:
+		emitline(J, F, exp);
+		emit(J, F, OP_FALSE);
+		break;
+	case EXP_THIS:
+		emitline(J, F, exp);
+		emit(J, F, OP_THIS);
+		break;
 
 	case EXP_REGEXP:
+		emitline(J, F, exp);
 		emit(J, F, OP_NEWREGEXP);
-		emitraw(J, F, addstring(J, F, exp->string));
-		emitraw(J, F, exp->number);
+		emitarg(J, F, addstring(J, F, exp->string));
+		emitarg(J, F, exp->number);
 		break;
 
 	case EXP_OBJECT:
+		emitline(J, F, exp);
 		emit(J, F, OP_NEWOBJECT);
 		cobject(J, F, exp->a);
 		break;
 
 	case EXP_ARRAY:
+		emitline(J, F, exp);
 		emit(J, F, OP_NEWARRAY);
 		carray(J, F, exp->a);
 		break;
 
 	case EXP_FUN:
-		emitfunction(J, F, newfun(J, exp->a, exp->b, exp->c, 0, F->strict));
+		emitline(J, F, exp);
+		emitfunction(J, F, newfun(J, exp->line, exp->a, exp->b, exp->c, 0, F->strict));
 		break;
 
 	case EXP_IDENTIFIER:
+		emitline(J, F, exp);
 		emitlocal(J, F, OP_GETLOCAL, OP_GETVAR, exp);
 		break;
 
 	case EXP_INDEX:
 		cexp(J, F, exp->a);
 		cexp(J, F, exp->b);
+		emitline(J, F, exp);
 		emit(J, F, OP_GETPROP);
 		break;
 
 	case EXP_MEMBER:
 		cexp(J, F, exp->a);
+		emitline(J, F, exp);
 		emitstring(J, F, OP_GETPROP_S, exp->b->string);
 		break;
 
@@ -605,28 +677,32 @@ static void cexp(JF, js_Ast *exp)
 	case EXP_NEW:
 		cexp(J, F, exp->a);
 		n = cargs(J, F, exp->b);
+		emitline(J, F, exp);
 		emit(J, F, OP_NEW);
-		emitraw(J, F, n);
+		emitarg(J, F, n);
 		break;
 
 	case EXP_DELETE:
-		cdelete(J, F, exp->a);
+		cdelete(J, F, exp);
 		break;
 
 	case EXP_PREINC:
 		cassignop1(J, F, exp->a);
+		emitline(J, F, exp);
 		emit(J, F, OP_INC);
 		cassignop2(J, F, exp->a, 0);
 		break;
 
 	case EXP_PREDEC:
 		cassignop1(J, F, exp->a);
+		emitline(J, F, exp);
 		emit(J, F, OP_DEC);
 		cassignop2(J, F, exp->a, 0);
 		break;
 
 	case EXP_POSTINC:
 		cassignop1(J, F, exp->a);
+		emitline(J, F, exp);
 		emit(J, F, OP_POSTINC);
 		cassignop2(J, F, exp->a, 1);
 		emit(J, F, OP_POP);
@@ -634,6 +710,7 @@ static void cexp(JF, js_Ast *exp)
 
 	case EXP_POSTDEC:
 		cassignop1(J, F, exp->a);
+		emitline(J, F, exp);
 		emit(J, F, OP_POSTDEC);
 		cassignop2(J, F, exp->a, 1);
 		emit(J, F, OP_POP);
@@ -641,11 +718,12 @@ static void cexp(JF, js_Ast *exp)
 
 	case EXP_VOID:
 		cexp(J, F, exp->a);
+		emitline(J, F, exp);
 		emit(J, F, OP_POP);
 		emit(J, F, OP_UNDEF);
 		break;
 
-	case EXP_TYPEOF: ctypeof(J, F, exp->a); break;
+	case EXP_TYPEOF: ctypeof(J, F, exp); break;
 	case EXP_POS: cunary(J, F, exp, OP_POS); break;
 	case EXP_NEG: cunary(J, F, exp, OP_NEG); break;
 	case EXP_BITNOT: cunary(J, F, exp, OP_BITNOT); break;
@@ -688,12 +766,14 @@ static void cexp(JF, js_Ast *exp)
 
 	case EXP_COMMA:
 		cexp(J, F, exp->a);
+		emitline(J, F, exp);
 		emit(J, F, OP_POP);
 		cexp(J, F, exp->b);
 		break;
 
 	case EXP_LOGOR:
 		cexp(J, F, exp->a);
+		emitline(J, F, exp);
 		emit(J, F, OP_DUP);
 		end = emitjump(J, F, OP_JTRUE);
 		emit(J, F, OP_POP);
@@ -703,6 +783,7 @@ static void cexp(JF, js_Ast *exp)
 
 	case EXP_LOGAND:
 		cexp(J, F, exp->a);
+		emitline(J, F, exp);
 		emit(J, F, OP_DUP);
 		end = emitjump(J, F, OP_JFALSE);
 		emit(J, F, OP_POP);
@@ -712,6 +793,7 @@ static void cexp(JF, js_Ast *exp)
 
 	case EXP_COND:
 		cexp(J, F, exp->a);
+		emitline(J, F, exp);
 		then = emitjump(J, F, OP_JTRUE);
 		cexp(J, F, exp->c);
 		end = emitjump(J, F, OP_JUMP);
@@ -820,12 +902,16 @@ static void cexit(JF, enum js_AstType T, js_Ast *node, js_Ast *target)
 	do {
 		prev = node, node = node->parent;
 		switch (node->type) {
-		default: /* impossible */ break;
+		default:
+			/* impossible */
+			break;
 		case STM_WITH:
+			emitline(J, F, node);
 			emit(J, F, OP_ENDWITH);
 			break;
 		case STM_FOR_IN:
 		case STM_FOR_IN_VAR:
+			emitline(J, F, node);
 			/* pop the iterator if leaving the loop */
 			if (F->script) {
 				if (T == STM_RETURN || T == STM_BREAK || (T == STM_CONTINUE && target != node)) {
@@ -846,6 +932,7 @@ static void cexit(JF, enum js_AstType T, js_Ast *node, js_Ast *target)
 			}
 			break;
 		case STM_TRY:
+			emitline(J, F, node);
 			/* came from try block */
 			if (prev == node->a) {
 				emit(J, F, OP_ENDTRY);
@@ -897,6 +984,7 @@ static void ctrycatch(JF, js_Ast *trystm, js_Ast *catchvar, js_Ast *catchstm)
 			if (!strcmp(catchvar->string, "eval"))
 				jsC_error(J, catchvar, "redefining 'eval' is not allowed in strict mode");
 		}
+		emitline(J, F, catchvar);
 		emitstring(J, F, OP_CATCH, catchvar->string);
 		cstm(J, F, catchstm);
 		emit(J, F, OP_ENDCATCH);
@@ -928,9 +1016,11 @@ static void ctrycatchfinally(JF, js_Ast *trystm, js_Ast *catchvar, js_Ast *catch
 			if (!strcmp(catchvar->string, "eval"))
 				jsC_error(J, catchvar, "redefining 'eval' is not allowed in strict mode");
 		}
+		emitline(J, F, catchvar);
 		emitstring(J, F, OP_CATCH, catchvar->string);
 		cstm(J, F, catchstm);
 		emit(J, F, OP_ENDCATCH);
+		emit(J, F, OP_ENDTRY);
 		L3 = emitjump(J, F, OP_JUMP); /* skip past the try block to the finally block */
 	}
 	label(J, F, L1);
@@ -958,18 +1048,20 @@ static void cswitch(JF, js_Ast *ref, js_Ast *head)
 			def = clause;
 		} else {
 			cexp(J, F, clause->a);
+			emitline(J, F, clause);
 			clause->casejump = emitjump(J, F, OP_JCASE);
 		}
 	}
 	emit(J, F, OP_POP);
 	if (def) {
+		emitline(J, F, def);
 		def->casejump = emitjump(J, F, OP_JUMP);
 		end = 0;
 	} else {
 		end = emitjump(J, F, OP_JUMP);
 	}
 
-	/* emit the casue clause bodies */
+	/* emit the case clause bodies */
 	for (node = head; node; node = node->b) {
 		clause = node->a;
 		label(J, F, clause->casejump);
@@ -991,6 +1083,7 @@ static void cvarinit(JF, js_Ast *list)
 		js_Ast *var = list->a;
 		if (var->b) {
 			cexp(J, F, var->b);
+			emitline(J, F, var);
 			emitlocal(J, F, OP_SETLOCAL, OP_SETVAR, var->a);
 			emit(J, F, OP_POP);
 		}
@@ -1015,6 +1108,7 @@ static void cstm(JF, js_Ast *stm)
 
 	case STM_EMPTY:
 		if (F->script) {
+			emitline(J, F, stm);
 			emit(J, F, OP_POP);
 			emit(J, F, OP_UNDEF);
 		}
@@ -1027,14 +1121,17 @@ static void cstm(JF, js_Ast *stm)
 	case STM_IF:
 		if (stm->c) {
 			cexp(J, F, stm->a);
+			emitline(J, F, stm);
 			then = emitjump(J, F, OP_JTRUE);
 			cstm(J, F, stm->c);
+			emitline(J, F, stm);
 			end = emitjump(J, F, OP_JUMP);
 			label(J, F, then);
 			cstm(J, F, stm->b);
 			label(J, F, end);
 		} else {
 			cexp(J, F, stm->a);
+			emitline(J, F, stm);
 			end = emitjump(J, F, OP_JFALSE);
 			cstm(J, F, stm->b);
 			label(J, F, end);
@@ -1046,6 +1143,7 @@ static void cstm(JF, js_Ast *stm)
 		cstm(J, F, stm->a);
 		cont = here(J, F);
 		cexp(J, F, stm->b);
+		emitline(J, F, stm);
 		emitjumpto(J, F, OP_JTRUE, loop);
 		labeljumps(J, F, stm->jumps, here(J,F), cont);
 		break;
@@ -1053,8 +1151,10 @@ static void cstm(JF, js_Ast *stm)
 	case STM_WHILE:
 		loop = here(J, F);
 		cexp(J, F, stm->a);
+		emitline(J, F, stm);
 		end = emitjump(J, F, OP_JFALSE);
 		cstm(J, F, stm->b);
+		emitline(J, F, stm);
 		emitjumpto(J, F, OP_JUMP, loop);
 		label(J, F, end);
 		labeljumps(J, F, stm->jumps, here(J,F), loop);
@@ -1073,6 +1173,7 @@ static void cstm(JF, js_Ast *stm)
 		loop = here(J, F);
 		if (stm->b) {
 			cexp(J, F, stm->b);
+			emitline(J, F, stm);
 			end = emitjump(J, F, OP_JFALSE);
 		} else {
 			end = 0;
@@ -1083,6 +1184,7 @@ static void cstm(JF, js_Ast *stm)
 			cexp(J, F, stm->c);
 			emit(J, F, OP_POP);
 		}
+		emitline(J, F, stm);
 		emitjumpto(J, F, OP_JUMP, loop);
 		if (end)
 			label(J, F, end);
@@ -1092,9 +1194,11 @@ static void cstm(JF, js_Ast *stm)
 	case STM_FOR_IN:
 	case STM_FOR_IN_VAR:
 		cexp(J, F, stm->b);
+		emitline(J, F, stm);
 		emit(J, F, OP_ITERATOR);
 		loop = here(J, F);
 		{
+			emitline(J, F, stm);
 			emit(J, F, OP_NEXTITER);
 			end = emitjump(J, F, OP_JFALSE);
 			cassignforin(J, F, stm);
@@ -1105,6 +1209,7 @@ static void cstm(JF, js_Ast *stm)
 			} else {
 				cstm(J, F, stm->c);
 			}
+			emitline(J, F, stm);
 			emitjumpto(J, F, OP_JUMP, loop);
 		}
 		label(J, F, end);
@@ -1138,6 +1243,7 @@ static void cstm(JF, js_Ast *stm)
 				jsC_error(J, stm, "unlabelled break must be inside loop or switch");
 		}
 		cexit(J, F, STM_BREAK, stm, target);
+		emitline(J, F, stm);
 		addjump(J, F, STM_BREAK, target, emitjump(J, F, OP_JUMP));
 		break;
 
@@ -1153,6 +1259,7 @@ static void cstm(JF, js_Ast *stm)
 				jsC_error(J, stm, "continue must be inside loop");
 		}
 		cexit(J, F, STM_CONTINUE, stm, target);
+		emitline(J, F, stm);
 		addjump(J, F, STM_CONTINUE, target, emitjump(J, F, OP_JUMP));
 		break;
 
@@ -1165,25 +1272,32 @@ static void cstm(JF, js_Ast *stm)
 		if (!target)
 			jsC_error(J, stm, "return not in function");
 		cexit(J, F, STM_RETURN, stm, target);
+		emitline(J, F, stm);
 		emit(J, F, OP_RETURN);
 		break;
 
 	case STM_THROW:
 		cexp(J, F, stm->a);
+		emitline(J, F, stm);
 		emit(J, F, OP_THROW);
 		break;
 
 	case STM_WITH:
+		F->lightweight = 0;
 		if (F->strict)
 			jsC_error(J, stm->a, "'with' statements are not allowed in strict mode");
 		cexp(J, F, stm->a);
+		emitline(J, F, stm);
 		emit(J, F, OP_WITH);
 		cstm(J, F, stm->b);
+		emitline(J, F, stm);
 		emit(J, F, OP_ENDWITH);
 		break;
 
 	case STM_TRY:
+		emitline(J, F, stm);
 		if (stm->b && stm->c) {
+			F->lightweight = 0;
 			if (stm->d)
 				ctrycatchfinally(J, F, stm->a, stm->b, stm->c, stm->d);
 			else
@@ -1194,15 +1308,18 @@ static void cstm(JF, js_Ast *stm)
 		break;
 
 	case STM_DEBUGGER:
+		emitline(J, F, stm);
 		emit(J, F, OP_DEBUGGER);
 		break;
 
 	default:
 		if (F->script) {
+			emitline(J, F, stm);
 			emit(J, F, OP_POP);
 			cexp(J, F, stm);
 		} else {
 			cexp(J, F, stm);
+			emitline(J, F, stm);
 			emit(J, F, OP_POP);
 		}
 		break;
@@ -1217,41 +1334,6 @@ static void cstmlist(JF, js_Ast *list)
 	}
 }
 
-/* Analyze */
-
-static void analyze(JF, js_Ast *node)
-{
-	if (isfun(node->type)) {
-		F->lightweight = 0;
-		return; /* don't scan inner functions */
-	}
-
-	if (node->type == STM_WITH) {
-		F->lightweight = 0;
-	}
-
-	if (node->type == STM_TRY && node->c) {
-		F->lightweight = 0;
-	}
-
-	if (node->type == EXP_IDENTIFIER) {
-		if (!strcmp(node->string, "arguments")) {
-			F->lightweight = 0;
-			F->arguments = 1;
-		} else if (!strcmp(node->string, "eval")) {
-			/* eval may only be used as a direct function call */
-			if (!node->parent || node->parent->type != EXP_CALL || node->parent->a != node)
-				js_evalerror(J, "%s:%d: invalid use of 'eval'", J->filename, node->line);
-			F->lightweight = 0;
-		}
-	}
-
-	if (node->a) analyze(J, F, node->a);
-	if (node->b) analyze(J, F, node->b);
-	if (node->c) analyze(J, F, node->c);
-	if (node->d) analyze(J, F, node->d);
-}
-
 /* Declarations and programs */
 
 static int listlength(js_Ast *list)
@@ -1261,7 +1343,7 @@ static int listlength(js_Ast *list)
 	return n;
 }
 
-static void cparams(JF, js_Ast *list)
+static void cparams(JF, js_Ast *list, js_Ast *fname)
 {
 	F->numparams = listlength(list);
 	while (list) {
@@ -1273,15 +1355,20 @@ static void cparams(JF, js_Ast *list)
 
 static void cvardecs(JF, js_Ast *node)
 {
+	if (node->type == AST_LIST) {
+		while (node) {
+			cvardecs(J, F, node->a);
+			node = node->b;
+		}
+		return;
+	}
+
 	if (isfun(node->type))
 		return; /* stop at inner functions */
 
 	if (node->type == EXP_VAR) {
 		checkfutureword(J, F, node->a);
-		if (F->lightweight)
-			addlocal(J, F, node->a, 1);
-		else
-			emitstring(J, F, OP_DEFVAR, node->a->string);
+		addlocal(J, F, node->a, 1);
 	}
 
 	if (node->a) cvardecs(J, F, node->a);
@@ -1295,8 +1382,12 @@ static void cfundecs(JF, js_Ast *list)
 	while (list) {
 		js_Ast *stm = list->a;
 		if (stm->type == AST_FUNDEC) {
-			emitfunction(J, F, newfun(J, stm->a, stm->b, stm->c, 0, F->strict));
-			emitstring(J, F, OP_INITVAR, stm->a->string);
+			emitline(J, F, stm);
+			emitfunction(J, F, newfun(J, stm->line, stm->a, stm->b, stm->c, 0, F->strict));
+			emitline(J, F, stm);
+			emit(J, F, OP_SETLOCAL);
+			emitarg(J, F, addlocal(J, F, stm->a, 0));
+			emit(J, F, OP_POP);
 		}
 		list = list->b;
 	}
@@ -1310,33 +1401,30 @@ static void cfunbody(JF, js_Ast *name, js_Ast *params, js_Ast *body)
 	if (F->script)
 		F->lightweight = 0;
 
-	if (body)
-		analyze(J, F, body);
-
 	/* Check if first statement is 'use strict': */
 	if (body && body->type == AST_LIST && body->a && body->a->type == EXP_STRING)
 		if (!strcmp(body->a->string, "use strict"))
 			F->strict = 1;
 
-	cparams(J, F, params);
+	F->lastline = F->line;
 
-	if (name) {
-		checkfutureword(J, F, name);
-		emit(J, F, OP_CURRENT);
-		if (F->lightweight) {
-			addlocal(J, F, name, 0);
-			emit(J, F, OP_INITLOCAL);
-			emitraw(J, F, findlocal(J, F, name->string));
-		} else {
-			emitstring(J, F, OP_INITVAR, name->string);
-		}
-	}
+	cparams(J, F, params, name);
 
 	if (body) {
 		cvardecs(J, F, body);
 		cfundecs(J, F, body);
 	}
 
+	if (name) {
+		checkfutureword(J, F, name);
+		if (findlocal(J, F, name->string) < 0) {
+			emit(J, F, OP_CURRENT);
+			emit(J, F, OP_SETLOCAL);
+			emitarg(J, F, addlocal(J, F, name, 0));
+			emit(J, F, OP_POP);
+		}
+	}
+
 	if (F->script) {
 		emit(J, F, OP_UNDEF);
 		cstmlist(J, F, body);
@@ -1350,10 +1438,10 @@ static void cfunbody(JF, js_Ast *name, js_Ast *params, js_Ast *body)
 
 js_Function *jsC_compilefunction(js_State *J, js_Ast *prog)
 {
-	return newfun(J, prog->a, prog->b, prog->c, 0, J->default_strict);
+	return newfun(J, prog->line, prog->a, prog->b, prog->c, 0, J->default_strict);
 }
 
-js_Function *jsC_compile(js_State *J, js_Ast *prog)
+js_Function *jsC_compilescript(js_State *J, js_Ast *prog, int default_strict)
 {
-	return newfun(J, NULL, NULL, prog, 1, J->default_strict);
+	return newfun(J, prog ? prog->line : 0, NULL, NULL, prog, 1, default_strict);
 }
